# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a command line interface for printing dependency information."""


import argparse
import logging
import sys
from typing import List, Optional

from depinfo.application import DisplayApplication, DisplayFormat


logger = logging.getLogger()


MAX_DEPTH = 5


def parse_arguments(argv: Optional[List[str]] = None) -> argparse.Namespace:
    """Define and immediately parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Display a package's dependencies, platform, and Python "
        "information.",
    )
    parser.add_argument(
        "package_name",
        metavar="PACKAGE",
        help="The package's distribution name.",
    )
    default_build_tools = "conda,flit,hatch,mamba,pbr,pip,poetry,setuptools,wheel"
    parser.add_argument(
        "--build-tools",
        help=f"A comma separated list of Python package managers "
        f"(default {default_build_tools}).",
        default=default_build_tools,
    )
    default_max_depth = 1
    parser.add_argument(
        "-d",
        "--max-depth",
        type=int,
        help=f"The maximum desired depth of nested dependencies to show "
        f"(default {default_max_depth}). Should be >= 0 and <{MAX_DEPTH}.",
        default=default_max_depth,
    )
    parser.add_argument(
        "--markdown",
        action="store_true",
        help="Display information as markdown tables (default false).",
    )
    default_log_level = "WARNING"
    parser.add_argument(
        "-l",
        "--log-level",
        help=f"The desired log level (default {default_log_level}).",
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        default=default_log_level,
    )
    return parser.parse_args(argv)


def main(argv: Optional[List[str]] = None) -> None:
    """Coordinate argument parsing, input validation, and program execution."""
    args = parse_arguments(argv)
    logging.basicConfig(level=args.log_level, format="[%(levelname)s] %(message)s")
    if not (0 <= args.max_depth < MAX_DEPTH):
        logger.critical(f"The maximum depth must be >=0 and <{MAX_DEPTH}.")
        sys.exit(2)
    if args.markdown:
        display_format = DisplayFormat.Markdown
    else:
        display_format = DisplayFormat.Simple
    DisplayApplication.run(
        package_name=args.package_name,
        display_format=display_format,
        build_tools=[token.strip() for token in args.build_tools.split(",")],
        max_depth=args.max_depth,
    )



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a display service registry."""


from typing import Type

from .abstract_display_service import AbstractDisplayService
from .display_format import DisplayFormat


class DisplayServiceRegistry:
    """Define a registry that returns display service classes based on given type."""

    @classmethod
    def display_service(
        cls, display_format: DisplayFormat
    ) -> Type[AbstractDisplayService]:
        """Return a display service class based on the given display format."""
        if display_format is DisplayFormat.Simple:
            from depinfo.infrastructure.application import SimpleDisplayService

            return SimpleDisplayService
        elif display_format is DisplayFormat.Markdown:
            from depinfo.infrastructure.application import MarkdownTableDisplayService

            return MarkdownTableDisplayService
        else:
            raise ValueError(f"Unknown display format {display_format}.")



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a service that displays dependency information as markdown tables."""


from operator import itemgetter
from typing import List, Tuple

from depinfo.application import AbstractDisplayService
from depinfo.domain import DependencyReport


class MarkdownTableDisplayService(AbstractDisplayService):
    """Define a service that displays dependency information as markdown tables."""

    @classmethod
    def display(cls, report: DependencyReport, max_depth: int = 1, **kwargs) -> None:
        """
        Display a dependency report to a desired maximum depth as markdown tables.

        Args:
            report: A dependency report instance.
            max_depth:  The maximum desired depth (default 1).
            **kwargs: Keyword arguments are ignored.

        """
        print(
            "\n".join(
                [
                    "",
                    "### Platform Information",
                    "",
                    *cls._format_table(
                        ["", ""],
                        [
                            (report.platform.name, report.platform.version),
                            (report.python.name, report.python.version),
                        ],
                    ),
                ]
            )
        )
        requirements = sorted(
            report.iter_unique_requirements(
                missing_version="**missing**", max_depth=max_depth
            ),
            key=itemgetter(0),
        )
        print(
            "\n".join(
                [
                    "",
                    "### Dependency Information",
                    "",
                    *cls._format_table(["Package", "Version"], requirements),
                ]
            )
        )
        tools = sorted(
            (
                (pkg.name, pkg.version)
                for pkg in report.build_tools
                if pkg.version is not None
            ),
            key=itemgetter(0),
        )
        print(
            "\n".join(
                [
                    "",
                    "### Build Tools Information",
                    "",
                    *cls._format_table(["Package", "Version"], tools),
                ]
            )
        )

    @classmethod
    def _format_table(
        cls, header: List[str], pairs: List[Tuple[str, str]]
    ) -> List[str]:
        """Format pairs of information as a markdown table with two columns."""
        if not pairs:
            return []
        max_len_name = max(max((len(pair[0]) for pair in pairs)), len(header[0]))
        max_len_version = max(max((len(pair[1]) for pair in pairs)), len(header[1]))
        result = [
            f"| {header[0]:^{max_len_name}} | {header[1]:^{max_len_version}} |",
            f"|:{'-' * max_len_name}-|-{'-' * max_len_version}:|",
        ]
        result.extend(
            f"| {name:<{max_len_name}} | {version:>{max_len_version}} |"
            for name, version in pairs
        )
        return result



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a platform description."""


from __future__ import annotations

import platform
from typing import NamedTuple


class Platform(NamedTuple):
    """
    Define a platform description.

    Attributes:
        name: The platform name, typically one of Linux, Darwin, or Windows.
        version: The platform's full release string and machine architecture.

    """

    name: str
    version: str

    @classmethod
    def create(cls) -> Platform:
        """Return a platform instance generated from the current environment."""
        return cls(
            name=platform.system(), version=f"{platform.release()}-{platform.machine()}"
        )



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a dependency report model with requirements information."""


from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, List, Tuple

from .package import Package
from .platform import Platform
from .python import Python


@dataclass(frozen=True)
class DependencyReport:
    """
    Define a dependency report with requirements information.

    A package is defined by its name, version, and requirements. It can be constructed
    from its name via a convenient factory method.

    Attributes:
        root: The root package instance for which to follow dependencies.
        platform: A `Platform` information instance.
        python: A `Python` information instance.
        build_tools: A list of package instances which are considered build tools rather
            than direct dependencies.
        packages: A map from package names to instances.

    """

    root: Package
    platform: Platform
    python: Python
    build_tools: List[Package]
    packages: Dict[str, Package]

    @classmethod
    def from_root(
        cls,
        root: str,
        build_tools: Iterable[str],
        max_depth: int = 1,
    ) -> DependencyReport:
        """
        Return a package instance potentially with its requirements.

        A factory class method that returns a package with its name and versions, as
        well as nested requirements as packages themselves.

        Args:
            root: The distribution name of the root package.
            build_tools: A list of build packages to include.
            max_depth: The maximum desired depth of requirements nesting.

        Returns:
            A dependency report instance with potentially nested requirements.

        """
        discovered = deque([(0, root)])
        packages: Dict[str, Package] = {}
        while len(discovered) > 0:
            level, name = discovered.popleft()
            if name in packages:
                continue
            packages[name] = pkg = Package.from_name(name)
            if level < max_depth:
                discovered.extend(((level + 1, req) for req in pkg.requirements))
        tools: List[Package] = []
        for name in build_tools:
            if name in packages:
                tools.append(packages[name])
                continue
            packages[name] = pkg = Package.from_name(name)
            tools.append(pkg)
        return cls(
            root=packages[root],
            build_tools=tools,
            packages=packages,
            platform=Platform.create(),
            python=Python.create(),
        )

    def iter_requirements(self, max_depth: int = 1) -> Iterator[Tuple[int, Package]]:
        """
        Iterate over the root package's nested requirements up to a maximum depth.

        Args:
            max_depth: The maximum desired depth of requirements nesting to iterate
                over.

        Yields:
            Dependency nesting level, package pairs.

        """
        discovered = deque([(0, self.root)])
        while len(discovered) > 0:
            level, pkg = discovered.popleft()
            if level < max_depth:
                discovered.extend(
                    ((level + 1, self.packages[req]) for req in pkg.requirements)
                )
            yield level, pkg

    def iter_unique_requirements(
        self, missing_version: str = "missing", max_depth: int = 1
    ) -> Iterator[Tuple[str, str]]:
        """
        Iterate over package name, version pairs up to a maximum dependency depth.

        Args:
            missing_version: A string to replace missing version information.
            max_depth: The maximum desired depth of dependency information.

        Yields:
            Package name, version pairs.

        """
        seen = set()
        for _, pkg in self.iter_requirements(max_depth=max_depth):
            if pkg.name in seen:
                continue
            yield pkg.name, missing_version if pkg.version is None else pkg.version
            seen.add(pkg.name)



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a service that displays dependency information as simple tables."""


from operator import itemgetter
from typing import List, Tuple

from depinfo.application import AbstractDisplayService
from depinfo.domain import DependencyReport


class SimpleDisplayService(AbstractDisplayService):
    """Define a service that displays dependency information as simple tables."""

    @classmethod
    def display(cls, report: DependencyReport, max_depth: int = 1, **kwargs) -> None:
        """
        Display a dependency report to a desired maximum depth as simple tables.

        Args:
            report: A dependency report instance.
            max_depth:  The maximum desired depth (default 1).
            **kwargs: Keyword arguments are ignored.

        """
        print(
            "\n".join(
                [
                    "",
                    "Platform Information",
                    "--------------------",
                    *cls._format_pairs(
                        [
                            (report.platform.name, report.platform.version),
                            (report.python.name, report.python.version),
                        ]
                    ),
                ]
            )
        )
        requirements = sorted(
            report.iter_unique_requirements(
                missing_version="missing", max_depth=max_depth
            ),
            key=itemgetter(0),
        )
        print(
            "\n".join(
                [
                    "",
                    "Dependency Information",
                    "----------------------",
                    *cls._format_pairs(requirements),
                ]
            )
        )
        print(
            "\n".join(
                [
                    "",
                    "Build Tools Information",
                    "-----------------------",
                    *cls._format_pairs(
                        sorted(
                            (
                                (pkg.name, pkg.version)
                                for pkg in report.build_tools
                                if pkg.version is not None
                            ),
                            key=itemgetter(0),
                        )
                    ),
                ]
            )
        )

    @classmethod
    def _format_pairs(cls, pairs: List[Tuple[str, str]]) -> List[str]:
        """Format pairs as two fixed width, left- and right-aligned columns."""
        if not pairs:
            return []
        max_len_name = max((len(pair[0]) for pair in pairs))
        max_len_version = max((len(pair[1]) for pair in pairs))
        return [
            f"{name:<{max_len_name}} {version:>{max_len_version}}"
            for name, version in pairs
        ]



# Copyright (c) 2018, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide deprecated functions for compatibility with previous versions."""


from warnings import warn

from .display_application import DisplayApplication


def print_dependencies(package_name: str) -> None:
    """Display dependency information for the given package in a simple format."""
    warn(
        DeprecationWarning(
            "The function 'print_dependencies' is deprecated since version 2.0.0. "
            "Please use either the command "
            "line interface with command 'depinfo' or import "
            "'depinfo.application.DisplayApplication' in your Python project."
        )
    )
    DisplayApplication.run(package_name=package_name)


def show_versions() -> None:
    """Display dependency information for this package in a simple format."""
    warn(
        DeprecationWarning(
            "The function 'show_versions' is deprecated since version 2.0.0. "
            "Please use either the command "
            "line interface with command 'depinfo' or import "
            "'depinfo.application.DisplayApplication' in your Python project."
        )
    )
    DisplayApplication.run(package_name="depinfo")



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide an application that displays dependency information."""


from typing import Iterable

from depinfo.domain import DependencyReport

from .display_format import DisplayFormat
from .display_service_registry import DisplayServiceRegistry


class DisplayApplication:
    """Define an application that displays dependency information."""

    @classmethod
    def run(
        cls,
        package_name: str,
        display_format: DisplayFormat = DisplayFormat.Simple,
        build_tools: Iterable[str] = (
            "conda",
            "flit",
            "hatch",
            "mamba",
            "pbr",
            "pip",
            "poetry",
            "setuptools",
            "wheel",
        ),
        max_depth: int = 1,
    ) -> None:
        """
        Display the given package's dependencies in the desired format.

        Args:
            package_name: The package name for which to generate dependency information.
            display_format: One of the supported display formats.
            build_tools: A list of build packages to include.
            max_depth: The maximum desired depth of requirements nesting.

        """
        report = DependencyReport.from_root(
            root=package_name,
            build_tools=build_tools,
            max_depth=max_depth,
        )
        DisplayServiceRegistry.display_service(display_format=display_format).display(
            report=report, max_depth=max_depth
        )



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a package model."""


from __future__ import annotations

import re
import sys
from dataclasses import dataclass
from typing import ClassVar, List, Optional, Pattern


if sys.version_info < (3, 8):
    from importlib_metadata import PackageNotFoundError, distribution
else:
    from importlib.metadata import PackageNotFoundError, distribution


@dataclass(frozen=True)
class Package:
    """
    Define a package model.

    A package is defined by its name, version, and requirements. It can be constructed
    from its name via a convenient factory method.

    Attributes:
        name: The package name.
        version: The package version.
        requirements: The package's requirements as other packages (if any).

    """

    name: str
    version: Optional[str]
    requirements: List[str]

    _req_pattern: ClassVar[Pattern] = re.compile(r"[\s();<>=]")

    @classmethod
    def from_name(cls, name: str) -> Package:
        """
        Return a package instance from its distribution name.

        A factory class method that returns a package with its name and version, as
        well as requirements.

        Args:
            name: A package's distribution name.

        Returns:
            A package instance with its version and requirements if it is installed in
            the current environment; otherwise only the name is set whereas the version
            and requirements are empty.

        """
        name = cls._normalize_name(name)
        try:
            dist = distribution(name)
        except PackageNotFoundError:
            result = cls(name=name, version=None, requirements=[])
        else:
            result = cls(
                name=name,
                version=dist.version,
                requirements=[]
                if dist.requires is None
                else [cls._normalize_name(cls._get_name(req)) for req in dist.requires],
            )
        return result

    @classmethod
    def _normalize_name(cls, name: str) -> str:
        """Normalize a package's name to lower case with hyphens only."""
        return name.lower().replace("_", "-")

    @classmethod
    def _get_name(cls, requirement: str) -> str:
        """
        Return the package name from requirement metadata.

        Args:
            requirement: Package requirement metadata as described in PEP 566
                (https://peps.python.org/pep-0566/).

        Returns:
            The package name.

        """
        return cls._req_pattern.split(requirement, maxsplit=1)[0]



# Copyright (c) 2022, Moritz E. Beber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide an abstract base class for dependency information display services."""


from abc import ABC, abstractmethod

from depinfo.domain import DependencyReport


class AbstractDisplayService(ABC):
    """Define an abstract base class for dependency information display services."""

    @classmethod
    @abstractmethod
    def display(cls, report: DependencyReport, max_depth: int = 1, **kwargs) -> None:
        """
        Display a dependency report to a desired maximum depth.

        Args:
            report: A dependency report instance.
            max_depth:  The maximum desired depth (default 1).
            **kwargs: Keyword arguments are passed on to the actual display method.

        """
